package PPD;

use strict;
use warnings;

use Exporter 'import';
our @EXPORT = qw(define_group define_ui_group declare get set maybe_set option generate);

### PPD data model ###

# Key => {
#	Key => key,
#	Name => name,		# descriptive name (default: same as Key)
#	Mandatory => bool,
#	Group => name,
#
#	# For main keywords:
#	Type => type,		# q=quoted, i=invocation, s=string (unquoted), b=boolean
#	Value => string,
#
#	# For option keywords:
#	Type => 'o',
#	Choice => choice,	# for UI options: PickOne / PickMany / Boolean
#	JCL => 1,		# if this is a JCL option
#	Section => sec,		# OrderDependency section: ExitServer / Prolog / DocumentSetup /
#				# PageSetup / JCLSetup (default if JCL=1) / AnySetup (default if JCL=0)
#	Priority => pri,	# OrderDependency priority, default = 100
#	Values => [		# possible values (mandatory)
#		{ Key => key,		# key (mandatory)
#		  Name => key,		# descriptive name (default: same as Key)
#		  PS => string,		# PS code to emit (mandatory)
#		  Default => 1,		# if this is the default value
#		},
#	]
# }
my %keywords = ();

# Key => {
#	Key => key,
#	UI => 1,		# if a UI group
#	Name => name,		# descriptive name
#	Keywords => [...]	# list of keywords inside the group
# }
my %groups = ();
my @head_groups = ();
my @ui_groups = ();
my @nonui_groups = ();

sub define_group($$) {
	my ($key, $g) = @_;
	!defined $groups{$key} or die "Group $key already exists\n";
	$groups{$key} = $g;
	$g->{Key} = $key;
	if ($g->{UI}) { push @ui_groups, $key; }
	elsif ($g->{Head}) { push @head_groups, $key; }
	else { push @nonui_groups, $key; }
}

sub define_head_group($$) {
	my ($key, $g) = @_;
	$g->{Head} = 1;
	define_group($key, $g);
}

sub define_ui_group($$) {
	my ($key, $g) = @_;
	$g->{UI} = 1;
	define_group($key, $g);
}

# declare("group", ["keyword", "type[!]", "value"], ...)
sub declare($@) {
	my $gn = shift @_;
	my $g = $groups{$gn} or die "Group $gn does not exist\n";
	for my $ktn (@_) {
		my ($key, $type, $value) = @$ktn;
		my $mand = 0;
		if ($type =~ s/!$//) { $mand = 1; }
		!$keywords{$key} or die "Keyword $key already declared\n";
		$keywords{$key} = {
			Key => $key,
			Name => $key,
			Group => $gn,
			Type => $type,
			Value => $value,
		};
		$keywords{$key}{Mandatory} = 1 if $mand;
		push @{$g->{Keywords}}, $keywords{$key};
	}
}

# get("keyword")
sub get($) {
	my ($key) = @_;
	my $kw = $keywords{$key};
	if ($kw) {
		return $kw->{'Value'};
	} else {
		return;
	}
}

# set("option", value) [must be already declared]
sub set($$) {
	my ($key, $value) = @_;
	my $k = $keywords{$key} or die "Unknown keyword $key (missing declaration)\n";
	$k->{Value} = $value;
}

# maybe_set("option", value) [must be already declared]
sub maybe_set($$) {
	my ($key, $value) = @_;
	my $k = $keywords{$key} or die "Unknown keyword $key (missing declaration)\n";
	$k->{Value} //= $value;
}

# option({ Group => group, Key => key, Name => name, ... other keyword parameters })
sub option($) {
	my ($o) = @_;
	my $key = $o->{Key};
	!defined $keywords{$key} or die "Keyword $key already known\n";
	my $g = $groups{$o->{Group}} or die "Keyword $key specified with an unknown group\n";
	$o->{Type} = 'o';
	$keywords{$key} = $o;
	push @{$g->{Keywords}}, $o;
}

### Known main keywords ###

# Group "f": file description
define_head_group('f', { Name => 'File version' });
declare('f',
	[ 'FormatVersion',		'q!',	'4.3' ],
	[ 'LanguageEncoding',		's!',	'ISOLatin1' ],
	[ 'LanguageVersion',		's!',	'English' ],
	[ 'FileVersion',		's!',	undef ],
	[ 'PCFileName',			's!',	undef ],
);

# Group "p": product description
define_head_group('p', { Name => 'Product description' });
declare('p',
	[ 'Manufacturer',		'q!',	undef ],
	[ 'ModelName',			'q!',	undef ],	# default: copy Product
	[ 'NickName',			'q!',	undef ],	# default: copy Product
	[ 'ShortNickName',		'q!',	undef ],	# default: copy NickName
	[ 'Product',			'q!',	undef ],
	[ 'PSVersion',			'q!',	undef ],
);

# Group "d": device capabilities
define_head_group('d', { Name => 'Device capabilities' });
declare('d',
	[ 'ColorDevice',		'b',	0 ],
	[ 'DefaultColorSpace',		's',	'Gray' ],
	[ 'FileSystem',			'b',	0 ],
	[ 'Extensions',			's',	undef ],
	[ 'FaxSupport',			's',	undef ],
	[ 'LanguageLevel',		'q',	undef ],
	[ 'Throughput',			'q',	undef ],
	[ 'TTRasterizer',		's',	undef ],
	[ '1284Modes',			's',	undef ],
	[ '1284DeviceID',		'q',	undef ],
	[ 'FreeVM',			'q',	undef ],
	# VMOption not supported yet
	[ 'VariablePaperSize',		'b',	1 ],
	[ 'Protocols',			's',	undef ],
);

# Group "c": CUPS options
define_head_group('c', { Name => 'CUPS options' });
declare('c',
	[ 'cupsProtocol',		's',	undef ],
);

# Group "j": JCL options
define_head_group('j', { Name => 'JCL options' });
declare('j',
	[ 'JCLBegin',			'q',	undef ],
	[ 'JCLToPSInterpreter',		'q',	undef ],
	[ 'JCLEnd',			'q',	undef ],
);

### Formatting of PPD ###

sub format_value($$) {
	my ($type, $value) = @_;
	$type //= 'q';
	if ($type eq 'q' || $type eq 'i') {
		# Quoted value / Invocation value
		if ($value =~ /["\x00-\x08\x0b-\x1f\x7f-\xff]/) { die "Invalid value: $value\n"; }
		$value = '"' . $value . '"';
		if ($value =~ /\n/) { $value .= "\n*End"; }
		return $value;
	} elsif ($type eq 's') {
		# String value
		return $value;
	} elsif ($type eq 'b') {
		# Boolean value
		return $value ? "True" : "False";
	} else {
		die "Unknown type '$type'\n";
	}
}

sub heading($) {
	my ($h) = @_;
	print "\n";
	print "%* ", "=" x length $h, "\n";
	print "%* ", $h, "\n";
	print "%* ", "=" x length $h, "\n";
	print "\n";
}

sub get_default($) {
	my ($o) = @_;
	for my $val (@{$o->{Values}}) {
		return $val->{Key} if $val->{Default};
	}
	return $o->{Values}->[0]->{Key};
}

sub emit_option($) {
	my ($o) = @_;
	my $key = $o->{Key};
	print "%* === $key ===\n";
	my $jcl = ($o->{JCL} // ($key =~ /^JCL/)) ? "JCL" : "";
	if (defined $o->{Choice}) {
		print "*${jcl}OpenUI *$key/", ($o->{Name} // $o->{Key}), ": ", $o->{Choice}, "\n";
		my $pri = $o->{Priority} // 100;
		my $sec = $o->{Section} // ($jcl ? "JCLSetup" : "AnySetup");
		print '*', ((defined $o->{UI}) ? "" : 'NonUI'), "OrderDependency $pri $sec *$key\n";
	}
	print "*Default$key: ", get_default($o), "\n";
	for my $v (@{$o->{Values}}) {
		my $ps = $v->{PS};
		# $ps = "\n$ps" unless $ps =~ /^\n/s;
		# $ps = "$ps\n" unless $ps =~ /\n$/s;
		print "*$key ", $v->{Key}, '/', ($v->{Name} // $v->{Key}), ': ', format_value('i', $ps), "\n";
	}
	if (defined $o->{Choice}) { print "*${jcl}CloseUI *$key\n"; }
	print "\n";
}

sub emit_keyword($) {
	my ($k) = @_;
	if ($k->{Type} eq 'o') {
		emit_option($k);
	} elsif (defined $k->{Value}) {
		print "*", $k->{Key}, ": ", format_value($k->{'Type'}, $k->{'Value'}), "\n";
	}
}

sub emit_group($) {
	my ($g) = @_;
	my $keywords = $g->{Keywords};
	$keywords && @$keywords or return;
	heading($g->{Name});
	if ($g->{UI}) { print '*OpenGroup: ', $g->{Key}, '/', ($g->{Name} // $g->{Key}), "\n\n"; }
	for my $k (@$keywords) {
		emit_keyword($k);
	}
	if ($g->{UI}) { print "\n", '*CloseGroup: ', $g->{Key}, "\n"; }
}

sub fill_defaults() {
	my $model = get('ModelName');
	if (defined $model) {
		maybe_set('Product', "($model)");
		maybe_set('NickName', $model);
	}

	my $nick = get('NickName');
	if ($nick) {
		maybe_set('ShortNickName', $nick);
	}

	my $psver = get('PSVersion');
	if (defined($psver) && $psver =~ /\((\d)/) {
		maybe_set('LanguageLevel', $1);
	}
}

sub check_missing() {
	for my $key (keys %keywords) {
		my $k = $keywords{$key};
		if ($k->{Mandatory} && !defined $k->{Value}) {
			die "Mandatory keyword $key not set\n";
		}
	}
}

sub generate() {
	fill_defaults();
	check_missing();

	print "*PPD-Adobe: ", get('FormatVersion'), "\n";
	print "*% PPD file generated by UCW PPD generator\n";

	for my $g (@head_groups, @ui_groups, @nonui_groups) {
		emit_group($groups{$g});
	}
}

42;
